#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
:Mod: dereferencer

:Synopsis:

:Author:
    servilla

:Created:
    1/26/23
"""
import copy

import daiquiri
from lxml import etree


logger = daiquiri.getLogger(__name__)


class Dereferencer(object):
    """
    Expands EML XML content by dereferencing "references" element to content defined
    by the "id" attribute of a source element.
    """

    def __init__(self, pretty_print=False):
        """
        Class init method.
        :param pretty_print: Boolean to indicate if dereferenced EML XML is formatted for viewing
        """
        self.pretty_print = pretty_print

    def dereference(self, xml: str) -> str:
        """
        Dereferences an EML XML document instance.
        :param xml: EML XML document instance as a unicode string.
        :return str: Expanded EML XML.
        """

        xml = xml.encode("utf-8")
        root = etree.fromstring(xml)

        references_nodes = root.xpath(".//references")
        for references in references_nodes:
            source_node = root.xpath(f".//*[@id='{references.text.strip()}']")[0]
            source_children = source_node.getchildren()
            parent_node = references.getparent()
            parent_node.remove(references)
            n_children = len(source_children)
            # Add children in reverse to ensure correct order when using E.insert() at position 0
            for child in range(n_children, 0, -1):
                replicant = copy.deepcopy(source_children[child - 1])
                parent_node.insert(0, replicant)

        return etree.tostring(root, pretty_print=self.pretty_print).decode("utf-8")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
:Mod: exceptions

:Synopsis:

:Author:
    servilla

:Created:
    1/21/23
"""
import daiquiri


logger = daiquiri.getLogger(__name__)


class EMLVPError(Exception):
    pass


class CircularReferenceIdError(EMLVPError):
    pass


class CustomUnitError(EMLVPError):
    pass


class DuplicateIdError(EMLVPError):
    pass


class InconsistentSystemError(EMLVPError):
    pass


class InspectionError(EMLVPError):
    pass


class MissingAdditionalMetadataDescribesIdError(EMLVPError):
    pass


class MissingAnnotationParentIdError(EMLVPError):
    pass


class MissingAnnotationReferencsIdError(EMLVPError):
    pass


class MissingReferenceIdError(EMLVPError):
    pass


class ParseError(EMLVPError):
    pass


class SaxError(EMLVPError):
    pass


class SchemaIncludeError(EMLVPError):
    pass


class ValidationError(EMLVPError):
    pass


class XPathError(EMLVPError):
    pass


class XMLSchemaError(EMLVPError):
    pass


class XMLSchemaParseError(EMLVPError):
    pass


class XMLSyntaxError(EMLVPError):
    pass


class XSLTError(EMLVPError):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
:Mod: parser

:Synopsis:

:Author:
    servilla

:Created:
    1/22/23
"""
import daiquiri
from lxml import etree

import emlvp.exceptions as exceptions


logger = daiquiri.getLogger(__name__)


class Parser(object):
    """
    Parses an EML XML document instance inspecting for non-schema related issues. See here for possible
    issues: https://eml.ecoinformatics.org/validation-and-content-references.html
    """

    def __init__(self, fail_fast: bool = False):
        """
        Class init method.
        :param fail_fast: Boolean to indicate whether parsing should fail immediately
        """
        self.fail_fast = fail_fast

    def parse(self, xml: str):
        """
        Parses an EML XML document instance inspecting for non-schema related issues.
        :param xml: EML XML document instance as a unicode string
        :return: None
        :raises emlvp.exceptions.ParseError: Raises ParseError on any invalid content found
        """

        xml = xml.encode("utf-8")
        root = etree.fromstring(xml)

        msg_queue = ""

        # Inspect id attributes in all elements to ensure uniqueness
        id_nodes = root.findall(".//*[@id]")
        ids = [i.attrib["id"] for i in id_nodes]
        id_duplicates = set([x for x in ids if ids.count(x) > 1])
        if len(id_duplicates) > 0:
            msg_duplicate_id = f"Duplicate id(s): {list(id_duplicates)}\n"
            msg_queue += msg_duplicate_id
            logger.debug(msg_duplicate_id)
            if self.fail_fast:
                raise exceptions.ParseError(msg_duplicate_id)

        # Remove duplicates
        ids = set(ids)

        # Inspect references elements for subject ids
        references_nodes = root.findall(".//{*}references")
        references = [r.text.strip() for r in references_nodes]
        missing_reference_id = False
        references_without_ids = []
        for r in references:
            if r not in ids:
                references_without_ids.append(r)
                missing_reference_id = True
        if missing_reference_id:
            msg_reference = f"Missing references id(s): {references_without_ids}\n"
            msg_queue += msg_reference
            logger.debug(msg_reference)
            if self.fail_fast:
                raise exceptions.ParseError(msg_reference)

        # Inspect for circular references (reference parent elements with id attributes)
        has_circular_reference = False
        circular_references = []
        for r in references_nodes:
            p = r.getparent()
            if p is not None and "id" in p.attrib:
                circular_references.append(f"{p.tag}::{r.text.strip()}")
                has_circular_reference = True
        if has_circular_reference:
            msg_circular_references = f"Circular references: {circular_references}\n"
            msg_queue += msg_circular_references
            logger.debug(msg_circular_references)
            if self.fail_fast:
                raise exceptions.ParseError(msg_circular_references)

        # Inspect for system attribute consistency
        has_system_inconsistency = False
        inconsistent_systems = []
        for r in references_nodes:
            r_system = None
            if "system" in r.attrib:
                r_system = r.attrib["system"]
            for i in id_nodes:
                if i.attrib["id"] == r.text.strip():
                    i_system = None
                    if "system" in i.attrib:
                        i_system = i.attrib["system"]
                    if r_system != i_system:
                        has_system_inconsistency = True
                        inconsistent_systems.append(f"{i.tag}::{r.text.strip()}")
                    break
        if has_system_inconsistency:
            msg_system_inconsistency = "Inconsistent system attribute(s): {inconsistent_systems}\n"
            msg_queue += msg_system_inconsistency
            logger.debug(msg_system_inconsistency)
            if self.fail_fast:
                raise exceptions.ParseError(msg_system_inconsistency)

        # Inspect custom units for STMML definitions
        has_undefined_custom_unit = False
        missing_custom_unit_ids = []
        custom_unit_nodes = root.findall(".//{*}customUnit")
        custom_units = set([u.text.strip() for u in custom_unit_nodes])
        unit_nodes = root.xpath("./additionalMetadata/metadata//*[local-name()='unitList']/*[local-name()='unit'][@id]")
        unit_ids = [i.attrib["id"] for i in unit_nodes]
        for custom_unit in custom_units:
            if custom_unit not in unit_ids:
                has_undefined_custom_unit = True
                missing_custom_unit_ids.append(custom_unit)
        if has_undefined_custom_unit:
            msg_undefined_custom_unit = f"Missing custom unit id(s): {missing_custom_unit_ids}\n"
            msg_queue += msg_undefined_custom_unit
            logger.debug(msg_undefined_custom_unit)
            if self.fail_fast:
                raise exceptions.ParseError(msg_undefined_custom_unit)

        # Inspect parents of annotation elements for subject id (sans the annotations element)
        has_missing_annotation_id = False
        missing_annotation_ids = []
        parents_nodes = root.xpath(".//*[local-name() != 'annotations']/annotation/parent::*")
        for p in parents_nodes:
            if "id" not in p.attrib:
                has_missing_annotation_id = True
                missing_annotation_ids.append(p.tag)
        if has_missing_annotation_id:
            msg_missing_annotation_id = f"Missing subject id for annotation parent(s): {missing_annotation_ids}\n"
            msg_queue += msg_missing_annotation_id
            logger.debug(msg_missing_annotation_id)
            if self.fail_fast:
                raise exceptions.ParseError(msg_missing_annotation_id)

        # Inspect references attribute of annotation(s) for subject id
        has_missing_annotation_references_id = False
        missing_annotation_references_ids = []
        annotation_nodes = root.xpath(".//annotations/annotation[@references]")
        annotation_references = set([r.attrib["references"] for r in annotation_nodes])
        for annotation_reference in annotation_references:
            if annotation_reference not in ids:
                has_missing_annotation_references_id = True
                missing_annotation_references_ids.append(annotation_reference)
        if has_missing_annotation_references_id:
            msg_missing_annotation_references = f"Missing subject id for annotation references: {missing_annotation_references_ids}\n"
            msg_queue += msg_missing_annotation_references
            logger.debug(msg_missing_annotation_references)
            if self.fail_fast:
                raise exceptions.ParseError(msg_missing_annotation_references)

        # Inspect additionalMetadata describes for subject id
        has_missing_describes_id = False
        missing_describes_ids = []
        describes_nodes = root.xpath(".//additionalMetadata/describes")
        describes = set([d.text.strip() for d in describes_nodes])
        for describe in describes:
            if describe not in ids:
                has_missing_describes_id = True
                missing_describes_ids.append(describe)
        if has_missing_describes_id:
            msg_missing_describes_id = f"Missing additionalMetadata describes subject id: {missing_describes_ids}\n"
            msg_queue += msg_missing_describes_id
            logger.debug(msg_missing_describes_id)
            if self.fail_fast:
                raise exceptions.ParseError(msg_missing_describes_id)

        # Fail slow
        if len(msg_queue) > 0:
            raise exceptions.ParseError(msg_queue.strip())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
:Mod: emlvp_cli

:Synopsis:

:Author:
    servilla

:Created:
    1/27/23
"""
import logging
from pathlib import Path
import sys

import click
import daiquiri

from emlvp.dereferencer import Dereferencer
from emlvp.exceptions import EMLVPError, ValidationError, ParseError
from emlvp.parser import Parser
from emlvp.validator import Validator


cwd = Path(".").resolve().as_posix()
logfile = cwd + "/emlvp.log"
daiquiri.setup(level=logging.INFO,
               outputs=(daiquiri.output.File(logfile), "stdout",))
logger = daiquiri.getLogger(__name__)


class Style(object):
    RED = '\033[31m'
    GREEN = '\033[32m'
    BLUE = '\033[34m'
    RESET = '\033[0m'


def vpd(xml: str, dereference: bool, fail_fast: bool, pretty_print: bool) -> str:
    """
    Validate, parse, and dereference EML XML file(s)
    :param xml:
    :param dereference:
    :param fail_fast:
    :param pretty_print:
    :return: None
    """

    p = Path(__file__).resolve().parent.as_posix()

    if "https://eml.ecoinformatics.org/eml-2.2.0" in xml:
        schema = p + "/schemas/EML2.2.0/xsd/eml.xsd"
    elif "eml://ecoinformatics.org/eml-2.1.1" in xml:
        schema = p + "/schemas/EML2.1.1/eml.xsd"
    elif "eml://ecoinformatics.org/eml-2.1.0" in xml:
        schema = p + "/schemas/EML2.1.0/eml.xsd"
    else:
        raise ValueError("Cannot determine schema")

    v = Validator(schema)
    v.validate(xml)
    p = Parser(fail_fast=fail_fast)
    p.parse(xml)
    if dereference:
        d = Dereferencer(pretty_print=pretty_print)
        xml = d.dereference(xml)
        v.validate(xml)
        p.parse(xml)

    return xml


def process_one_document(doc: str, dereference: bool, fail_fast: bool, pretty_print: bool, verbose: int):
    with open(doc, "r") as f:
        xml = f.read()
        try:
            xml = vpd(xml, dereference, fail_fast, pretty_print)
            if verbose >= 1:
                print(f"{doc}\n")
                if verbose >= 2:
                    print(xml)
        except (ValidationError, ParseError) as e:
            if verbose >= 0:
                print(f"{doc}\n{Style.RED}{e}{Style.RESET}\n")
                if verbose >= 2:
                    print(xml)
            raise EMLVPError(e)
    pass


help_target = "Either EML XML file or directory containing EML XML file(s)."
help_dereference = "Dereference EML XML file(s) (default is False)."
help_fail_fast = "Exit on first exception encountered (default is False)."
help_pretty_print = "Pretty print output for dereferenced EML XML (default is False)."
help_statistics = "Show post processing inspection statistics."
verbose_help = "Send output to standard out (-v or -vv or -vvv for increasing output)."


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("target", required=True, nargs=-1)
@click.option("-d", "--dereference", is_flag=True, default=False, help=help_dereference)
@click.option("-f", "--fail-fast", is_flag=True, default=False, help=help_fail_fast)
@click.option("-p", "--pretty-print", is_flag=True, default=False, help=help_pretty_print)
@click.option("-s", "--statistics", is_flag=True, default=False, help=help_statistics)
@click.option("-v", "--verbose", count=True, help=verbose_help)
def main(target: tuple, dereference: bool, fail_fast: bool, pretty_print: bool, statistics: bool, verbose: int):
    """
        Performs validation of EML XML file(s)\n
            1. XML schema validation\n
            2. EML parsing for references/id resolution\n
            3. Dereference references/id into expanded EML XML and re-validate/parse\n

        \b
            TARGET: EML XML file or directory containing EML XML file(s) (may be repeated)
    """
    docs_processed = 0
    docs_with_exceptions = 0

    for t in target:
        if Path(t).is_file():
            try:
                docs_processed += 1
                process_one_document(doc=t, dereference=dereference, fail_fast=fail_fast,
                                     pretty_print=pretty_print, verbose=verbose)
            except EMLVPError:
                docs_with_exceptions += 1
        elif Path(t).is_dir():
            for tf in Path(t).glob("*.xml"):
                try:
                    docs_processed += 1
                    process_one_document(doc=str(tf), dereference=dereference, fail_fast=fail_fast,
                                         pretty_print=pretty_print, verbose=verbose)
                except EMLVPError:
                    docs_with_exceptions += 1
        else:
            logger.error(f"Target {t} is not a file or directory")
            sys.exit(1)

        if statistics:
            print(f"Total documents validated: {docs_processed}")
            print(f"Documents that failed validation: {docs_with_exceptions}")

    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: setup.py

:Synopsis:

:Author:
    servilla

:Created:
    2023-01-21
"""
from os import path
from setuptools import find_packages, setup


here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(here, "LICENSE"), encoding="utf-8") as f:
    full_license = f.read()

with open(path.join(here, "./src/emlvp/VERSION.txt"), encoding="utf-8") as f:
    version = f.read()


setup(
    name='emlvp',
    version=version,
    description='EMLvp (Ecological Metadata Lanaguage validator and parser)',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Mark Servilla",
    url="https://github.com/servilla/EMLvp",
    license=full_license,
    packages=find_packages(where="src", include=["emlvp"]),
    package_dir={"": "src"},
    include_package_data=True,
    python_requires=" >= 3.10",
    install_requires=["lxml>=4.9.2"],
    entry_points={"console_scripts": ["emlvp=emlvp.emlvp_cli:main"]},
    classifiers=["License :: OSI Approved :: Apache Software License",],
)


def main():
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
:Mod: validator

:Synopsis:

:Author:
    servilla

:Created:
    1/21/23
"""
import os
from pathlib import Path

import daiquiri
from lxml import etree

import emlvp.exceptions as exceptions


logger = daiquiri.getLogger(__name__)


def schema_path() -> str:
    """
    Return path to installed schemas.
    :return: Path to emlvp installed schemas
    :rtype: str
    """
    return os.path.abspath(os.path.dirname(__file__)) + "/schemas"


class Validator(object):
    """
    Validates an EML XML document for being well formed and schema syntax correct.
    """

    def __init__(self, schema: str):
        """
        Class init method.
        :param schema: path to root schema eml.xsd
        """
        self.schema = schema
        if not Path(self.schema).is_file():
            msg = f"Cannot locate root schema file: {schema}"
            raise IOError(msg)

    def validate(self, xml: str):
        """
        Validates an EML XML document instance
        :param xml: EML XML document instance as a unicode string
        :return: None
        :raises emlvp.exceptions.ValidationError: Raises ValidationError on any invalid content found
        """

        xml = xml.encode("utf-8")

        try:
            doc = etree.fromstring(xml)
            schema = etree.XMLSchema(file=self.schema)
            schema.assertValid(doc)
        except etree.DocumentInvalid as e:
            logger.debug(e)
            raise exceptions.ValidationError(e)
        except etree.ParserError as e:
            logger.debug(e)
            raise exceptions.ValidationError(e)
        except etree.XIncludeError as e:
            logger.debug(e)
            raise exceptions.ValidationError(e)
        except etree.XMLSchemaParseError as e:
            logger.debug(e)
            raise exceptions.ValidationError(e)
        except etree.XMLSyntaxError as e:
            logger.debug(e)
            raise exceptions.ValidationError(e)

